// -*- C++ -*-

#ifndef	_pixrect_pixrect_h
#define	_pixrect_pixrect_h

#include <unistd.h>
#include <std.h>
#include <sys/types.h>
#include <stdlib.h>

// const int replacements for original pixrect macros

const int  PIX_ERR		= -1;
const int  PIX_SRC		= 0xC << 1;
const int  PIX_DST		= 0xA << 1;
const int  PIX_CLR		= 0x0 << 1;
const int  PIX_SET		= 0xF << 1;
const int  PIX_DONTCLIP		= 0x1;
const int  PIX_CLIP		= 0x0;
const int  PR_FORCE_UPDATE	= 1 << 24;
const int  PR_DONT_DEGAMMA	= 1 << 23;
const int  PR_DEGAMMA		= 1 << 22;

// inline replacements for original pixrect macros

inline int  PIX_NOT(int op)	      {return (op ^ 0x1E);}
inline int  PIXOP_NEEDS_DST(int op)   {return((op>>1)^op) & PIX_NOT(PIX_DST);}
inline int  PIXOP_NEEDS_SRC(int op)   {return((op>>2)^op) & PIX_NOT(PIX_SRC);}
inline int  PIX_COLOR(int c)	      {return c<<5;}
inline int  PIX_OPCOLOR(int op)	      {return op>>5;}
inline int  PIX_OP_CLIP(int op)	      {return op & 0x1f;}
inline int  PIX_OP(int op)	      {return op & 0x1e;}
inline int  ROUNDUP(int val,int gran) {return((val) - 1 | (gran) - 1) + 1;}

#ifndef pr_product
#if defined(__sun__) && !defined(__sparc__)
#define	pr_product(a, b)	((short)(a) * (short)(b))
#else
#define	pr_product(a, b)	((a) * (b))
#endif
#endif

// structure used to specify fields in a 32-bit pixel 
union fbunit 
{
  unsigned int packed;	       /* whole-sale deal */
  struct 
  {
    unsigned int A:8;	       /* unused, for now */
    unsigned int B:8;	       /* blue channel */
    unsigned int G:8;	       /* green channel */
    unsigned int R:8;	       /* red channel */
  } channel; 		       /* access per channel */
};

// underlying low-level pixrect functions used by Pixobj

extern "C" 
{
  struct pixrectops;

  struct pr_pos { 
    int	x, y;
  };
  
  struct pr_size {
    int	x, y;
  };
  
  typedef struct pixrect 
  {
    struct	pixrectops *pr_ops;	/* operations appropriate to this pr */
    struct	pr_size pr_size;	/* pixels per dimension */
    int	pr_depth;		/* bits per pixel */
    caddr_t	pr_data;		/* device-dependent data */
  } Pixrect;
  
  struct pixrectops 
  {
    int	(*pro_rop)();
    int	(*pro_stencil)();
    int	(*pro_batchrop)();
    int	(*pro_nop)();		/* place holder */
    int	(*pro_destroy)();
    int	(*pro_get)();
    int	(*pro_put)();
    int	(*pro_vector)();
    Pixrect * (*pro_region)();
    int	(*pro_putcolormap)();
    int	(*pro_getcolormap)();
    int	(*pro_putattributes)();
    int	(*pro_getattributes)();
  };	
  
  Pixrect *pr_open();
  int pr_replrop();
  int pro_line();
  
  struct pr_prpos 
  {
    Pixrect*	pr;
    pr_pos   	pos;
  };	
  
  struct pr_subregion 
  {
    Pixrect*	pr;
    pr_pos  	pos;
    pr_size  	size;
  };
  
} // extern "C"

#include <pixrect/memvar.h>
#include <pixrect/pr_line.h>
#include <pixrect/pr_io.h>
#include <pixrect/pixfont.h>

class Pixobj 
{
public:

  // Constructors:
  // 
  // 1. Pixobj screen;
  // 2. Pixobj window(screen, 100, 100, 800, 600);
  // 3. Pixobj mem(100, 100, 1);
  // 4. Pixobj po(pixrect);
  // 5. Pixobj po("raster.file");
  
  Pixobj() : tex(NULL), pr(reg.pr),pos(reg.pos),size(reg.size) { 
      pr = pr_open("/dev/fb"); init(); }

  Pixobj(Pixobj& as, int ax, int ay, int aw, int ah) : tex(NULL), pr(reg.pr),pos(reg.pos), size(reg.size) { 
      pr = (*(as.pr)->pr_ops->pro_region)(as.pr, ax, ay, aw, ah); init(); }
  
  Pixobj(int aw, int ah, int ad, short* data = NULL) : tex(NULL), pr(reg.pr),pos(reg.pos), size(reg.size) { 
      if(data) pr=mem_point(aw,ah,ad,data);else pr=mem_create(aw,ah,ad);init(); }
  
  Pixobj(Pixrect* p) : tex(NULL), pr(reg.pr),pos(reg.pos), size(reg.size) { 
      pr = (*(p)->pr_ops->pro_region)(p, 0, 0, p->pr_size.x, p->pr_size.y); init(); }

  Pixobj(char* n)  : tex(NULL), pr(reg.pr),pos(reg.pos), size(reg.size) { 
      FILE *fp = fopen(n, "r");  colormap_t colormap; pr = pr_load(fp, &colormap); fclose(fp); init();}

  // Destructor
  
   
  ~Pixobj() { (*pr_ops.pro_destroy)(pr); }

  // Initialize values (private)
  
  void init() { size.x=prWidth(); size.y=prHeight(); pos.x = pos.y = x = y = 0; pr_ops = *(pr->pr_ops); }
  
  // Size and depth

  int prWidth()	 { return pr->pr_size.x; }
  int prHeight() { return pr->pr_size.y; }
  int prDepth()	 { return pr->pr_depth; }

  // Set drawing attributes for the Pixobj

  Pixobj& X(int ax)    { x = ax; return *this; }
  Pixobj& Y(int ay)    { y = ay; return *this; }	
  Pixobj& Op(int ao)   { op &= ~0x1ff; op |= ao; return *this;}
  Pixobj& Color(int c) { op &= 0x1ff; op |= PIX_COLOR(c); return *this;}
  Pixobj& LineWidth(int w) 	{ brush.width = w; return *this; }
  
  // Set multiple attributes

  Pixobj& Pos(int ax, int ay) 	{ pos.x = ax; pos.y = ay; return *this; }
  Pixobj& Size(int ax, int ay) 	{ size.x = ax; size.y = ay; return *this; }
  
  Pixobj& Rect(int ax, int ay, int aw, int ah)	
  { pos.x=ax; pos.y=ay; size.x=aw; size.y=ah; return *this; }
  
  // Variants of pr_get
  
  int Get()		  { return (*pr_ops.pro_get)(pr, x, y);}
  int Get(int ax, int ay) { return (*pr_ops.pro_get)(pr, x=ax, y=ay);}
    
  // Variants of pr_put
  
  Pixobj& Put(int c)		   { (*pr_ops.pro_put)(pr, x, y, c);}
  Pixobj& Put(int ax,int ay,int c) { (*pr_ops.pro_put)(pr, x=ax, y=ay, c);}

  // Vector:
  //
  // 1. p.Vector(10, 10); -- draws a vector from x, y to 10, 10 -> x=10 y=10
  // 2. p.Vector(0, 0, 10, 10); -- draws a vector from 0, 0 to 10, 10

  Pixobj& Vector(int ax, int ay)			
  {(*pr_ops.pro_vector)(pr, x, y, ax, ay, op, 0); x=ax; y=ay; return *this;}
	
  Pixobj& Vector(int ax, int ay, int ax1, int ay1)
  { (*(pr)->pr_ops->pro_vector)(pr, ax, ay, ax1, ay1, op, 0);return *this;}
  
  Pixobj& Line(int ax, int ay)
  { pro_line(pr, x, y, ax, ay, &brush, tex, op); x=ax; y=ay; return *this;}
  
  Pixobj& Line(int ax, int ay, int ax1, int ay1)
  { pro_line(pr, ax, ay, ax1, ay1, &brush, tex, op); return *this;}

  // Operators
  
  Pixobj& operator = (Pixobj& s) {(*pr_ops.pro_rop)(pr, pos.x, pos.y, size.x, size.y, op, s.pr, s.x, s.y); return *this;}
  
  Pixobj& operator = (int ao) {(*pr_ops.pro_rop)(pr, pos.x, pos.y, size.x, size.y, ao, NULL, 0, 0);return *this;}

  Pixobj& operator << (int ao) {(*pr_ops.pro_rop)(pr, 0, 0, pr->pr_size.x, pr->pr_size.y, ao, NULL, 0, 0);return *this;}

  Pixobj& operator *= (Pixobj& s) {pr_replrop(pr, pos.x, pos.y, size.x, size.y, op, s.pr, s.x, s.y); return *this;}

  // Copies all pixels where mask != 0 from src to dest
  
  Pixobj& stencil(Pixobj& mask, Pixobj& src)
  {(*pr_ops.pro_stencil)(pr, pos.x, pos.y, size.x, size.y, op, mask.pr, mask.x, mask.y, src.pr, src.x, src.y); return *this;} 
  
  // Paints all pixels where mask != 0 to colorcode in mask.op
  
  Pixobj& stencil(Pixobj& mask)
  {(*pr_ops.pro_stencil)(pr, pos.x, pos.y, size.x, size.y, PIX_OPCOLOR(mask.op), mask.pr, mask.x, mask.y, NULL, 0, 0); return *this;}

  // Load a pixrect
  // (Obsolete)
  int Load(char *n) { FILE *fp = fopen(n, "r"); if (fp == NULL) return 1; colormap_t colormap; delete pr; pr = pr_load(fp, &colormap); fclose(fp); size.x = pr->pr_size.x; size.y=pr->pr_size.y; return 0;}

  int get_attr() {int a; (*pr_ops.pro_getattributes)(pr, &a); return a;}
  Pixobj& put_attr(int& a) {(*pr_ops.pro_putattributes)(pr, &a);}
  
  pr_subregion reg;
  Pr_brush brush;
  Pr_texture* tex;
  pixrectops pr_ops; // copy of pixrects function pointers for faster access
  
  // some references for easier access to substructures

  Pixrect*& pr;
  pr_pos& pos;
  pr_size& size;
  int op;
  int x;
  int y;
};

#endif	/* !_pixrect_pixrect_h */







